from kubernetes import client, config, watch
try:
    from kubernetes.client.rest import ApiException
except ImportError:
    from kubernetes.client.exceptions import ApiException
import os
import threading
import queue
import base64
import time
import json
import subprocess
import uuid
import logging
import sys


CA_SECRET_NAME = os.environ['CA_SECRET_NAME']
CA_SECRET_NAMESPACE = os.environ['CA_SECRET_NAMESPACE']

DEFAULT_LOG_FORMATTER = logging.Formatter(
    '%(asctime)s %(levelname)-8s %(name)s:%(lineno)s '
    '[%(threadName)s] - %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
fh_out = logging.StreamHandler(sys.stdout)
fh_out.setLevel(logging.INFO)
fh_out.setFormatter(DEFAULT_LOG_FORMATTER)
logger.addHandler(fh_out)


class NotReadyException(Exception):
    pass


def get_ca_certificate_data(coreV1Api: client.CoreV1Api) -> str:
    secret = coreV1Api.read_namespaced_secret(CA_SECRET_NAME,
                                              CA_SECRET_NAMESPACE,
                                              _request_timeout=5)
    crt = secret.data.get('tls.crt')
    crtString = base64.b64decode(crt).decode("utf-8")
    delimiter = "-----BEGIN CERTIFICATE-----"
    crtStringArr = crtString.split(delimiter)
    rootCrtString = delimiter + crtStringArr[-1]
    return rootCrtString


def get_ca_pem(obj: client.V1Secret, coreV1Api: client.CoreV1Api) -> str:
    logger.info("Inject pem truststore: %s %s" % (
    obj.metadata.name, obj.metadata.namespace))
    rootCrtString = get_ca_certificate_data(coreV1Api)
    rootCrt = base64.b64encode(bytes(rootCrtString, "utf-8")).decode("utf-8")
    return rootCrt


def get_ca_pkcs12(obj: client.V1Secret, coreV1Api: client.CoreV1Api) -> str:
    logger.info("Inject pkcs12 truststore: %s %s" % (
    obj.metadata.name, obj.metadata.namespace))
    rootCrt = get_ca_certificate_data(coreV1Api)
    ns = obj.metadata.namespace
    passwordSecretName = obj.metadata.annotations.get(
        'trust-manager/password-secret')
    try:
        passwordSecret = coreV1Api.read_namespaced_secret(passwordSecretName,
                                                          ns,
                                                          _request_timeout=5)
    except client.ApiException as ex:
        if (str(ex.status) == str(404)):
            raise NotReadyException("password {} is not ready".format(passwordSecretName))
    passwordEncoded = passwordSecret.data.get('password')
    password = base64.b64decode(passwordEncoded).decode("utf-8")
    id = str(uuid.uuid4().hex)
    with open("/tmp/{}.pem".format(id), "w") as f:
        f.write(rootCrt)
    logger.info("write /tmp/%s.pem" % id)
    command = "keytool -importcert -noprompt -keystore /tmp/{}.p12 -storepass {} -alias ca -file /tmp/{}.pem -storetype PKCS12 -providername SunJSSE".format(
        id, password, id)
    prog = subprocess.Popen(command.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    try:
        out, error = prog.communicate(timeout=15)
        logger.info("out: %s error: %s" % (out, error))
        if prog.returncode:
            raise NotReadyException(error)
    except subprocess.TimeoutExpired as e:
        prog.kill()
        raise NotReadyException(e)

    logger.info("reading /tmp/%s.p12" % id)
    try:
        with open("/tmp/{}.p12".format(id), "rb") as f:
            content = f.read()
            pkcs12 = base64.b64encode(content).decode("utf-8")
            return pkcs12
    except FileNotFoundError as e:
        raise NotReadyException(e)


def inject_truststore(obj: client.V1Secret, coreV1Api: client.CoreV1Api):
    annotations = obj.metadata.annotations
    if obj.data is None:
        obj.data = {}
    store_type = annotations.get('trust-manager/store-type')
    used_by = annotations.get('trust-manager/used-by')
    excludeCa = 'trust-manager/exclude-platform-ca' in annotations
    data = None
    keyname = None
    if store_type.lower() == "pkcs12":
        if obj.data.get('truststore.p12') is None:
            if not excludeCa:
                data = get_ca_pkcs12(obj, coreV1Api)
            else:
                data = ""
            keyname = "truststore.p12"
    elif store_type.lower() == "pem":
        if obj.data.get('ca.crt') is None:
            if not excludeCa:
                data = get_ca_pem(obj, coreV1Api)
            else:
                data = ""
            keyname = "ca.crt"

    if data != None and keyname:
        logger.info("Patch truststore: %s %s" % (obj.metadata.name, obj.metadata.namespace))
        jsonpatch = {'data': {
            keyname: data
        }}
        logger.info("Dumping patched truststore object: %s" % json.dumps(jsonpatch))
        coreV1Api.patch_namespaced_secret(obj.metadata.name,
                                          obj.metadata.namespace,
                                          body=jsonpatch)
    else:
        logger.info("Nothing to patch truststore: %s %s" % (obj.metadata.name, obj.metadata.namespace))


def wait_util_ca_ready(coreV1Api: client.CoreV1Api):
    # post-install-job sleeps 90s before creating ca secret
    time.sleep(90)
    while True:
        try:
            get_ca_certificate_data(coreV1Api)
            logger.info("ca secret is present")
            return
        except Exception as e:
            logger.warning("Warning: ca secret is not present")
            time.sleep(10)


def process(q: queue.Queue):
    v1 = client.CoreV1Api()
    errors = 0
    while True:
        try:
            item = q.get()
            logger.info("Processing: %s %s" % (item.metadata.name, item.metadata.namespace))
            secret = v1.read_namespaced_secret(item.metadata.name,
                                               item.metadata.namespace,
                                               _request_timeout=5)
            inject_truststore(secret, v1)
            logger.info("Done: %s %s" % (item.metadata.name, item.metadata.namespace))
            errors = 0
        except Exception as e:
            logger.exception("Processing Error")
            errors = errors + 1
            if (isinstance(e, NotReadyException)):
                logger.error("put back into queue (not ready): %s" % e)
                q.put(item)
            elif (isinstance(e, ApiException)):
                if (int(e.status) >= 500):
                    logger.error("put back into queue (k8s api server error): %s" % e)
                    q.put(item)
            if (errors % 10 == 0):
                # if we hit exceptions consistently, sleep a while before re-try
                time.sleep(30)


if __name__ == "__main__":
    logger.info("Starting truststore injector...")
    config.load_incluster_config()
    v1 = client.CoreV1Api()
    q = queue.Queue()

    # wait until ca secret is present
    wait_util_ca_ready(v1)

    # turn-on the worker thread
    threading.Thread(target=process, daemon=True, args=(q,)).start()

    while True:
        try:
            for event in watch.Watch().stream(v1.list_secret_for_all_namespaces, label_selector="trust-manager/entity=truststore", _request_timeout=180):
                obj = event['object']
                operation = event['type']
                labels = obj.metadata.labels
                logger.info("Receiving: %s %s" % (obj.metadata.name, operation))
                if "DELETED" != operation:
                    q.put(obj)
        except Exception as e:
            logger.warning("Watch is broken: %s" % e)
            continue
